Author : Anurag Mukkara (anurag_m@csail.mit.edu)
Date : 30 Jan 2016

Poolmalloc is a simpler wrapper over doug lea's malloc implementation(http://gee.cs.oswego.edu/dl/).
This was developed as part of Whirltool profiling tool described in Whirlpool:Improving Dynamic Cache Management
with Static Data Classification, Anurag Mukkara, Nathan Beckmann, Daniel Sanchez, ASPLOS-21, April 2016.
See Section 4.0, 4.1. 4.3 of the paper for a brief introduction into how poolMalloc fits into the whirltool framework.

Poolmalloc classifies memory allocation requests based on callpoint id and allocates pages in independent
memory regions called pools. Each memory pool is an mspace in doug lea's malloc implementation.
The malloc wrappers are heavily based on the implementation of google's tcmalloc(http://goog-perftools.sourceforge.net/doc/tcmalloc.html).

There are two ways in which poolMalloc can be used.
In profiling mode, it records the number of bytes allocated from each callpoint
and dumps this information at the end of program execution.
In execution mode, it reads a file containing callpoint id -> pool mappings(see usage section for input file format).
For each allocation request, it identifies the callpoint id and allocates pages from the corresponding pool.

Callpoints ids are either the calling PC or a hash of the stacktrace at the time of a memory alloction request.
For the latter, we use libunwind's backtrace functionality.

The implementation is thread-safe and uses very basic(and slow!) p_thread locks to prevent simulatenous modifications to critical data structures.
It has been mostly used with single threaded workloads, but it works fine with multithreaded workloads as well(but thread-safety hasn't been
heavily tested).

The code uses some hooks to interface with zsim simulator. But for native execution without zsim, all these hooks end up being NOP instructions.

Please read the code for detailed comments on how the code works.

Files:
dlmalloc.h/c
Doug Lea's malloc implementation, with slight modifications
that care of tagging and untagging pages to zsim/jigsaw VCs.

libc_override_gcc_and_weak.h
libc_override_glibc.h
libc_override.h
libc_override_redefine.h

Header files that override standard libc memory allocator.
These are basically stripped down versions of those used in google's
tcmalloc implementation

zsim-hooks.h
Hooks used to interface with zsim simulator.
If not using zsim, these are effectively nops.

poolMalloc.h/c
Main files that define wrappers and maintain state needed to identify callpoints, and track stats.

simpleHash.h
we produce each callpoint id by walking the stack and hashing the last two return PCs. The hash
function is implemented here.

unit_tests
Examples that show how to get stack traces.

Dependencies:
libunwind8

Usage
Build the library(libpoolmalloc.so)
    make

Set LD_PRELOAD=/PATH/TO/libpoolmalloc.so
Set mandatory environment vars like PM_MAPPINGSFILE, PM_OUTPUTFILE
Set optional environment vars like PM_MODE, PM_SIGNATURE, PM_TRACEDEPTH, PM_RELADDR

To get started with the library, run an application in profile mode and generate a file that maps each callpoint
to a separate pool. Then use this file as PM_MAPPINGSFILE to run the application in execution mode.

Format of PM_MAPPINGSFILE
numPools
numCallpoints
callpoint1 pool
callpoint2 pool
callpoint3 pool
...
...
callpointN pool
