#ifndef POOL_MALLOC_LIBC_OVERRIDE_REDEFINE_H_
#define POOL_MALLOC_LIBC_OVERRIDE_REDEFINE_H_

#include <new>

#ifndef __THROW    // I guess we're not on a glibc-like system
# define __THROW   // __THROW is just an optimization, so ok to make it ""
#endif

void* operator new(size_t size)                  { return pool_new(size);       }
void operator delete(void* p) __THROW            { pool_delete(p);              }
void* operator new[](size_t size)                { return pool_newarray(size);  }
void operator delete[](void* p) __THROW          { pool_deletearray(p);         }
void* operator new(size_t size, const std::nothrow_t& nt) __THROW {
  return pool_new_nothrow(size, nt);
}
void* operator new[](size_t size, const std::nothrow_t& nt) __THROW {
  return pool_newarray_nothrow(size, nt);
}
void operator delete(void* ptr, const std::nothrow_t& nt) __THROW {
  return pool_delete_nothrow(ptr, nt);
}
void operator delete[](void* ptr, const std::nothrow_t& nt) __THROW {
  return pool_deletearray_nothrow(ptr, nt);
}
extern "C" {
  void* malloc(size_t s) __THROW                 { return pool_malloc(s);       }
  void  free(void* p) __THROW                    { pool_free(p);                }
  void* realloc(void* p, size_t s) __THROW       { return pool_realloc(p, s);   }
  void* calloc(size_t n, size_t s) __THROW       { return pool_calloc(n, s);    }
  void  cfree(void* p) __THROW                   { pool_cfree(p);               }
  void* memalign(size_t a, size_t s) __THROW     { return pool_memalign(a, s);  }
  void* valloc(size_t s) __THROW                 { return pool_valloc(s);       }
  void* pvalloc(size_t s) __THROW                { return pool_pvalloc(s);      }
  int posix_memalign(void** r, size_t a, size_t s) __THROW {
    return pool_posix_memalign(r, a, s);
  }
  void malloc_stats(void) __THROW                { pool_malloc_stats();         }
  int mallopt(int cmd, int v) __THROW            { return pool_mallopt(cmd, v); }
#ifdef HAVE_STRUCT_MALLINFO
  struct mallinfo mallinfo(void) __THROW         { return pool_mallinfo();      }
#endif
  size_t malloc_size(void* p) __THROW            { return pool_malloc_size(p); }
  size_t malloc_usable_size(void* p) __THROW     { return pool_malloc_size(p); }
}  // extern "C"

// No need to do anything at tcmalloc-registration time: we do it all
// via overriding weak symbols (at link time).
static void ReplaceSystemAlloc() { }

#endif  // POOL_MALLOC_LIBC_OVERRIDE_REDEFINE_H_
