/** $lic$
 * Copyright (C) 2015-2016 by Massachusetts Institute of Technology
 *
 * This file is part of Whirltool.
 *
 * Whirltool is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * If you use this software in your research, we request that you reference
 * the Whirlpool paper ("Whirlpool: Improving Dynamic Cache Management with Static Data
 * Classification", Mukkara, Beckmann and Sanchez, ASPLOS-21, April 2016) as the
 * source in any publications that use this software, and that you send us a
 * citation of your work.
 *
 * Whirltool is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Whirltool.  If not, see <http://www.gnu.org/licenses/>.
 *
 **/

#ifndef __ZSIM_HOOKS_H__
#define __ZSIM_HOOKS_H__

#include <stdint.h>
#include <stdio.h>

//Avoid optimizing compilers moving code around this barrier
#define COMPILER_BARRIER() { __asm__ __volatile__("" ::: "memory");}

//These need to be in sync with the simulator
#define ZSIM_MAGIC_OP_ROI_BEGIN         (1025)
#define ZSIM_MAGIC_OP_ROI_END           (1026)
#define ZSIM_MAGIC_OP_REGISTER_THREAD   (1027)
#define ZSIM_MAGIC_OP_HEARTBEAT         (1028)
#define ZSIM_MAGIC_OP_WORK_BEGIN        (1029)
#define ZSIM_MAGIC_OP_WORK_END          (1030)
#define ZSIM_MAGIC_OP_FFI_FF_EXIT       (1050)
#define ZSIM_MAGIC_OP_CREATE_SHARE      (2034)
#define ZSIM_MAGIC_OP_TAG_ADDRESS       (2035)
#define ZSIM_MAGIC_OP_UNTAG_ADDRESS     (2036)
#define ZSIM_MAGIC_OP_SKIP_INSTR_BEGIN  (2037)
#define ZSIM_MAGIC_OP_SKIP_INSTR_END    (2038)
#define ZSIM_MAGIC_OP_RECORD_ALLOCATION (2039)

#ifdef __x86_64__
#define HOOKS_STR  "HOOKS"
static inline void zsim_magic_op(uint64_t op) {
    COMPILER_BARRIER();
    __asm__ __volatile__("xchg %%rcx, %%rcx;" : : "c"(op));
    COMPILER_BARRIER();
}
#else
#define HOOKS_STR  "NOP-HOOKS"
static inline void zsim_magic_op(uint64_t op) {
    //NOP
}
#endif

static inline void zsim_roi_begin() {
    printf("[" HOOKS_STR "] ROI begin\n");
    zsim_magic_op(ZSIM_MAGIC_OP_ROI_BEGIN);
}

static inline void zsim_roi_end() {
    zsim_magic_op(ZSIM_MAGIC_OP_ROI_END);
    printf("[" HOOKS_STR  "] ROI end\n");
}

static inline void zsim_heartbeat() {
    zsim_magic_op(ZSIM_MAGIC_OP_HEARTBEAT);
}

static inline void zsim_work_begin() { zsim_magic_op(ZSIM_MAGIC_OP_WORK_BEGIN); }
static inline void zsim_work_end() { zsim_magic_op(ZSIM_MAGIC_OP_WORK_END); }

static inline void zsim_ffi_ff_exit(){ zsim_magic_op(ZSIM_MAGIC_OP_FFI_FF_EXIT); }

static inline void zsim_skip_instr_begin() { zsim_magic_op(ZSIM_MAGIC_OP_SKIP_INSTR_BEGIN); }
static inline void zsim_skip_instr_end() { zsim_magic_op(ZSIM_MAGIC_OP_SKIP_INSTR_END); }

#define zsim_create_share(share)                            \
    __asm__ __volatile__("xchg %%rcx, %%rcx;"               \
                         : "=a"(share)                      \
                         : "c"(ZSIM_MAGIC_OP_CREATE_SHARE)  \
                         :"memory");

#define zsim_tag_address(startAddr, endAddr, share)         \
    __asm__ __volatile__("xchg %%rcx, %%rcx;"               \
                         :                                  \
                         : "S"(startAddr), "d"(endAddr), "b"(share), "c"(ZSIM_MAGIC_OP_TAG_ADDRESS) \
                         : "memory");

#define zsim_untag_address(startAddr, endAddr)              \
    __asm__ __volatile__("xchg %%rcx, %%rcx;"               \
                         :                                  \
                         : "S"(startAddr), "d"(endAddr), "c"(ZSIM_MAGIC_OP_UNTAG_ADDRESS) \
                         : "memory");

#define zsim_record_allocation(hash, size)                       \
    __asm__ __volatile__("xchg %%rcx, %%rcx;"                \
                         :                                  \
                         : "S"(hash), "d"(size), "c"(ZSIM_MAGIC_OP_RECORD_ALLOCATION)    \
                         : "memory");

#endif /*__ZSIM_HOOKS_H__*/
