#!/usr/bin/python
'''
/** $lic$
 * Copyright (C) 2015-2016 by Massachusetts Institute of Technology
 *
 * This file is part of Whirltool.
 *
 * Whirltool is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * If you use this software in your research, we request that you reference
 * the Whirlpool paper ("Whirlpool: Improving Dynamic Cache Management with Static Data
 * Classification", Mukkara, Beckmann and Sanchez, ASPLOS-21, April 2016) as the
 * source in any publications that use this software, and that you send us a
 * citation of your work.
 *
 * Whirltool is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Whirltool.  If not, see <http://www.gnu.org/licenses/>.
 *
 **/
'''

import argparse, os
import operator

# Input : A list of callpoints sorted by total allocation size
# Output: A subset of callpoints that account for (1 - tolerance) proportion of
#         total bytes allocated

# Used to preproces a full list of callpoints before collecting miss curves.
# PC and callpoint are used interchangeably

parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('infile', type=str, default='hashCount.in', help = 'input file containing list of hashes')
parser.add_argument('outfile', type=str, default='hashCount.out', help = 'file to dump sorted list of top hashes')
parser.add_argument('top', type=int, default=64, help = 'number of hashes to dump in outfile')
parser.add_argument('tolerance', type=float, default=0.001, help = 'fraction of allocation sizes to discard')
parser.add_argument('--printPC', action="store_true", default=False, help = 'Print list of PCs by total allocation sizes')

args = parser.parse_args()

# Input is sorted by allocation size
pcToAllocSize = {}
fractionToDiscard = args.tolerance

for line in open(args.infile):
    if len(line.strip().split(" ")) == 2:
        (pc,allocsize) = line.strip().split(" ")
        if pc in pcToAllocSize.keys():
            pcToAllocSize[pc] += int(allocsize) # Multiple allocations from same PC
        else:
            pcToAllocSize[pc] = int(allocsize)

pcToAllocSize = [(v, k) for k, v in pcToAllocSize.iteritems()]
pcToAllocSize.sort(key=lambda x: x[0], reverse=True) # Sort by sizes(first element)

if args.printPC:
    for (size, pc) in pcToAllocSize:
        print pc, ":", size

# Discard pcs with less than some fraction of allocations
totalAllocSize = sum([pair[0] for pair in pcToAllocSize])

runningSum = 0
i = 0
for i in range(len(pcToAllocSize)):
    runningSum += pcToAllocSize[i][0]
    if runningSum >= (1 - fractionToDiscard) * totalAllocSize:
        break

selectedPCs = min(args.top - 1, i + 1)

print "Selecting %d/%d PCs with a tolerance of %f" % (selectedPCs, len(pcToAllocSize), fractionToDiscard)

fp = open(args.outfile, 'w+')

numPools = selectedPCs
numPCs = selectedPCs

print >>fp, numPools
print >>fp, numPCs

for i in range(len(pcToAllocSize)):
    (allocsize, pc) = pcToAllocSize[i]
    print >>fp, pc + ' ' + str(i+1)
    if (i+1) == selectedPCs:
        break

print 'Done'
